#include <iostream>
#include <fstream>
#include <cstring>
#include <sstream>
#include <termios.h>
#include <unistd.h>
#include <openssl/evp.h>
#include "Vernam.h"
#include "csprng.h"
using namespace std;

void* getStream(const char* filename, ios_base::openmode mode) {
	if (filename == NULL || strcmp(filename, "-") == 0 || strcmp(filename, "") == 0)
		return mode == ifstream::in ? (void*)&cin : &cout;
	return new fstream(filename, mode | ifstream::binary);
}

void disableCinEcho() {
	termios tty;
	tcgetattr(STDIN_FILENO, &tty);
	/* we want to disable echo */
	tty.c_lflag &= ~ECHO;
	tcsetattr(STDIN_FILENO, TCSANOW, &tty);
}

void enableCinEcho() {
	termios tty;
	tcgetattr(STDIN_FILENO, &tty);
	/* we want to reenable echo */
	tty.c_lflag |= ECHO;
	tcsetattr(STDIN_FILENO, TCSANOW, &tty);
}

Vernam::Vernam(int argc, char* argv[]) {
	int processedArgs = 0;
	_arg0 = argv[0];
	_exitCode = 0;
	_mode = UNDEFINED;
	_cinFile = NULL;
	_coutFile = NULL;
	for (int i = 1; i < argc; ++i) {
		string arg = string(argv[i]);
		if (arg == "--help") {
			_mode = HELP;
			return;
		} else if (arg == "--encrypt") {
			_mode = ENCRYPT;
			++processedArgs;
		} else if (arg == "--decrypt") {
			_mode = DECRYPT;
			++processedArgs;
		} else if (arg.substr(0, 11) == "--password=") {
			_password = arg.substr(11);
			++processedArgs;
		} else if (arg.substr(0, 8) == "--input=") {
			_cinFile = arg.substr(8).c_str();
			++processedArgs;
		} else if (arg.substr(0, 9) == "--output=") {
			_coutFile = arg.substr(9).c_str();
			++processedArgs;
		} else if (arg.length() >= 2 && arg[0] == '-' && arg[1] != '-')
			splitSingleArgs(argc - i, argv + i, &processedArgs);
	}
	if (processedArgs < argc - 1)
		throw "Undefined commands. Try --help.";
}

void Vernam::splitSingleArgs(int argc, char* argv[], int* processedArgs) {
	int pwd = 0, in = 0, out = 0;
	int i = 1;
	int index = 0;
	char ch;
	while ((ch = argv[0][i++])) {
		switch (ch) {
			case 'h': _mode = HELP; return;
			case 'e': _mode = ENCRYPT; break;
			case 'd': _mode = DECRYPT; break;
			case 'p': pwd = ++index; break;
			case 'i': in  = ++index; break;
			case 'o': out = ++index; break;
			default:  
				char* msg = (char*)"Undefined command: \"- \". Try --help.";
				msg[22] = ch;
				throw msg;
		}
	}
	if (index >= argc)
		throw "Less parameters provided than expected. Try --help.";
	if (pwd > 0) _password = argv[pwd];
	if (in  > 0) _cinFile = argv[in];
	if (out > 0) _coutFile = argv[out];
	*processedArgs += index + 1;
}

void Vernam::run() {
	switch (_mode) {
		case HELP:
			showHelp(true);
			break;
		case ENCRYPT:
			initRun();
			encrypt();
			break;
		case DECRYPT:
			initRun();
			decrypt();
			break;
		default:
			showHelp(false);
			_exitCode = 2;
			break;
	}
}

void Vernam::initRun() {
	// verify password
	if (_password.empty()) {
		cerr << "Please enter a password: ";
		cerr.flush();
		disableCinEcho();
		string pwd1;
		if (_mode == ENCRYPT) {  // ask twice only for encryption
			getline(cin, pwd1);
			cerr << endl << "Please re-enter your password: ";
			cerr.flush();
		}
		getline(cin, _password);
		enableCinEcho();
		cerr << endl;
		cerr.flush();
		if (_mode == ENCRYPT) {  // ask twice only for encryption
			if (pwd1 != _password)
				throw "Your entered passwords are not equal.";
		}
	}
	
	// open streams
	_cin = (istream*)getStream(_cinFile, ifstream::in);
	_cout = (ostream*)getStream(_coutFile, ifstream::out);
}

/**
 * method requires *is >> noskipws
 **/
streamsize readBytes(istream* is, char* buf, streamsize bufLen) {
	streamsize result = 0;
	while (is->good() && result < bufLen)
		*is >> buf[result++];
	// if !is->good() one \0 too much has been read and result must be decreased
	return result == bufLen ? result : result - 1;
}

void Vernam::doEncrypt(char* salt) {
	const int passLen = _password.length();
	char* pass = new char[passLen];
	strcpy(pass, _password.c_str());
	
	// Hash salt and password and seed CSPRNG
	char* hash = new char[EVP_MAX_MD_SIZE];
	unsigned int hashLen;
	const EVP_MD* md = EVP_sha512();
	EVP_MD_CTX* mdctx = EVP_MD_CTX_create();
	EVP_DigestInit_ex(mdctx, md, NULL);
	EVP_DigestUpdate(mdctx, pass, passLen);
	EVP_DigestUpdate(mdctx, (unsigned char*)salt, SALT_LEN);
	EVP_DigestFinal_ex(mdctx, (unsigned char*)hash, &hashLen);
	EVP_MD_CTX_destroy(mdctx);	

	csprng* rand = new csprng(hash, hashLen);
	delete hash;
	delete pass;
	
	// encrypt data from _cin
	char buf[BUF_LEN];
	char rnd[BUF_LEN];
    *_cin >> noskipws;
	streamsize len;
	do {
		len = readBytes(_cin, buf, BUF_LEN);
		rand->getRandom(rnd, len);
		for (int i = 0; i < len; ++i) {
			buf[i] ^= rnd[i];
		}
		_cout->write(buf, len);
	} while(len == BUF_LEN);
	_cout->flush();
	delete rand;
}

void Vernam::encrypt() {
	*_cout << MAGIC_NUMBER;
	// Create and write salt
	char* salt = new char[SALT_LEN];
	csprng::generateSeed(salt, SALT_LEN);
	_cout->write(salt, SALT_LEN);
	doEncrypt(salt);
	delete salt;
}

void Vernam::decrypt() {
	char* magicNumber = new char[MAGIC_NUM_LEN];
	_cin->read(magicNumber, MAGIC_NUM_LEN);
	if (memcmp(magicNumber, MAGIC_NUMBER, MAGIC_NUM_LEN) != 0)
		throw "Input is not an encrypted Vernam data stream.";
	delete magicNumber;
	char* salt = new char[SALT_LEN];
	_cin->read(salt, SALT_LEN);
	doEncrypt(salt);
}

void Vernam::showHelp(bool showDetails) {
	cout << "This is Vernam " << VERSION << ". Nice to meet you." << endl;
    cout << "Usage: " << _arg0 << " [OPTION]..." << endl;
    if (!showDetails)
        cout << "Try `" << _arg0 << " --help` for more information." << endl;
    else
        cout << "Vernam encrypts or decrypts data.\n\nOptions:\n"
			 << "  -e, --encrypt            Encrypt data\n"
			 << "  -d, --decrypt            Decrypt data\n"
			 << "  -p, --password=PASSWORD  Use PASSWORD for encryption/decryption instead of\n"
			 << "                           reading from stdin.\n"
			 << "  -i, --input=FILE         Read data from FILE instead of stdin.\n"
			 << "  -o, --output=FILE        Write data to FILE instead of stdout.\n"
			 << endl
			 << "Examples:\n"
			 << "  " << _arg0 << " --encrypt --input=plain.txt --output=cipher.otp\n"
			 << "  " << _arg0 << " -eio plain.txt cipher.otp\n"
			 << "      Encrypts the file plain.txt and writes the encrypted output to\n"
			 << "      cipher.otp. Vernam will ask for a password at runtime.\n"
			 << endl
			 << "  " << _arg0 << " --decrypt --input=cipher.otp --output=plain.txt\n"
			 << "  " << _arg0 << " -dio cipher.otp plain.txt\n"
			 << "      Decrypts the file cipher.otp and writes the decrypted output to\n"
			 << "      plain.txt. Vernam will ask for the password at runtime.\n"
			 << "      If the password was wrong Vernam has no information to know that. Most\n"
			 << "      likely garbage will be decrypted instead.\n"
			 << endl
			 << "  " << _arg0 << " --encrypt --password=secret\n"
			 << "  " << _arg0 << " --encrypt --password=secret --input=- --output=-\n"
			 << "  " << _arg0 << " -ep secret\n"
			 << "  " << _arg0 << " -epio secret - -\n"
			 << "      Encrypts any data from stdin using the password \"secret\" and writes the\n"
			 << "      encrypted output to stdout. This might usually be used in a pipe chain\n"
			 << "      such as:\n"
			 << "          echo \"message\" | " << _arg0 << " -ep secret | mail -s\"subject\" foo@foo.bar\n"
			 << endl
			 << "  " << _arg0 << " --decrypt --password=secret --input=cipher.opt\n"
			 << "  " << _arg0 << " --decrypt --password=secret --input=cipher.opt --output=-\n"
			 << "  " << _arg0 << " -dpi secret cipher.otp\n"
			 << "  " << _arg0 << " -dpio secret cipher.otp -\n"
			 << "      Decrypts any data from the file cipher.otp using the password \"secret\"\n"
			 << "      and writes the decrypted output to stdout. This might be used to read\n"
			 << "      encrypted plain texts within a terminal or to be used in a pipe chain:\n"
			 << "          cat cipher.otp | " << _arg0 << " -dp secret\n"
			 << endl;
}
